
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "polygraph.h"

#include <iostream.h>
#include <iomanip.h>

#include "Rnd.h"
#include "base/ObjId.h"
#include "TblDistr.h"
#include "base/RndPermut.h"
#include "ContentCfg.h"
#include "ContentMgr.h"
#include "runtime/LogComment.h"

#include "ContentSym.h"
#include "ServerSym.h"

#include "ContentSel.h"

#include "gadgets.h"
#include "base/rndDistrStat.h"


ContentSel::ContentSel(): theDirAccSel(0) {
}

ContentSel::~ContentSel() {
	delete theDirAccSel;
}

// note: we are being fed "unique" configurations so
// cfg->host() does not mean much here
void ContentSel::configure(const ServerSym *cfg) {
	Array<ContentCfg*> ccfgs;

	// get "global" content information
	Array<ContentSym*> csyms;
	Array<double> cprobs;
	if (!cfg->contents(csyms, cprobs) || !csyms.count()) {
		cerr << cfg->loc() << "no content specs for server " << cfg->kind() << endl;
		exit(-2);
	}
	Assert(csyms.count() == cprobs.count());
	TheContentMgr.get(csyms, ccfgs);

	// get direct access info
	Array<ContentSym*> dasyms;
	if (!cfg->directAccess(dasyms) || !dasyms.count()) {
		cerr << cfg->loc() << "no direct_access specs for server " << cfg->kind() << endl;
		exit(-2);
	}
	TheContentMgr.get(dasyms, theDirectAccess);

	// check that all direct access cfgs are unique
	// and build cprobsViaDa map
	Array<double> cprobsViaDa;
	cprobsViaDa.stretch(theDirectAccess.count());
	for (int d = 0; d < theDirectAccess.count(); ++d) {
		ContentCfg *da = theDirectAccess[d];
		int count = 0;
		for (int c = 0; c < ccfgs.count(); ++c) {
			if (ccfgs[c]->id() == da->id()) {
				cprobsViaDa.append(cprobs[c]);
				count++;
			}
		}
		if (count != 1) {
			cerr << dasyms[d]->loc() 
				<< "direct_access spec below counted " << count
				<< " times in the global contents specs (must be 1):"
				<< endl << *dasyms[d] << endl;
			exit(-2);
		}
	}

	// compute direct access probabilities using known content distribution
	double daProb = 0;
	for (int d = 0; d < theDirectAccess.count(); ++d)
		daProb += cprobsViaDa[d];
	Array<double> daprobs(theDirectAccess.count());
	for (int d = 0; d < theDirectAccess.count(); ++d)
		daprobs.append(cprobsViaDa[d]/daProb); // d's contribution

	// make selector for direct access content
	theDirAccSel = TblDistr::FromDistrTable("da-sel", daprobs);
	Assert(theDirAccSel);

	// calculate expected global content distribution
	// and average cachability of pages
	Array<double> expCProbs(ccfgs.count());
	double expCProbsSum = 0;
	for (int c = 0; c < ccfgs.count(); ++c) {
		const ContentCfg *cc = ccfgs[c];

		// compute average number of oids of type c generated per da
		double contPerDa = 0;
		for (int d = 0; d < theDirectAccess.count(); ++d) {
			ContentCfg *da = theDirectAccess[d];
			//cerr << here << "da[" << d << "]*" << daprobs[d] << '/' << daProb << ": " << da->compContPerCall(cc) << endl;
			contPerDa += da->compContPerCall(cc) * daprobs[d];
		}
		
		// check that content type is accessible
		if (contPerDa <= 0) {
			cerr << csyms[c]->loc()
				<< ": the content description below is not accessible "
				<< "(is not in direct_access closure):"
				<< endl << *csyms[c] << endl;
			exit(-2);
		}

		expCProbs[c] = contPerDa;
		expCProbsSum += contPerDa;
	}

	// XXX: this should be moved to some kind of reportCfg() or similar
	static int repCount = 0;
	if (!repCount++) {

		// finilize calculation of expCProbs
		// and compute average cachability of pages
		double chbSum = 0;
		double sizeSum = 0;
		for (int c = 0; c < ccfgs.count(); ++c) {
			expCProbs[c] /= expCProbsSum;
			chbSum += expCProbs[c] * ccfgs[c]->theChbRatio;
			sizeSum += expCProbs[c] * RndDistrStat(ccfgs[c]->theSize).mean();
		}

		// report planned and expected global content distributions
		Comment(7) << "Content distribution on server " << cfg->kind() << ':' << endl;
		Comment << setw(15) << "content"
			<< ' ' << setw(15) << "planned%" 
			<< ' ' << setw(15) << "likely%" 
			<< ' ' << setw(15) << "error%" 
			<< endl;
		for (int c = 0; c < ccfgs.count(); ++c) {
			const ContentCfg *cc = ccfgs[c];
			Comment << setw(15) << (const char*)cc->kind()
				<< ' ' << setw(15) << 100*cprobs[c]
				<< ' ' << setw(15) << 100*expCProbs[c]
				<< ' ' << setw(15) << 100*Ratio(expCProbs[c]-cprobs[c], cprobs[c])
				<< endl;
		}
		Comment << "expected average cachability: " << 100*chbSum << '%' << endl;
		Comment << "expected average object size: " << sizeSum << "Bytes" << endl;
		Comment << endc;
	}
}

ContentCfg *ContentSel::getDir(const ObjId &oid) {
	theDirAccSel->rndGen()->seed(RndPermut(oid.hash(), rndContentSel));
	const int tid = (int)theDirAccSel->trial();
	Assert(tid >= 0 && tid < theDirectAccess.count());
	return theDirectAccess[tid];
}
