
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl.h"
#include "xparser.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <strstream.h>

#include "Assert.h"
#include "String.h"
#include "PglLexer.h"
#include "PglPp.h"
#include "PglParser.h" /* for tokens */


Array<String*> PglPp::TheDirs;


/* PglPp */

PglPp::PglPp(const String &fname): theDepth(0) {
	open(fname);
}

PglPp::~PglPp() {
	while (theDepth > 0) close();
}

LexToken PglPp::scan() {
	Assert(lexer());

	advance();
	while (!filtered() && theDepth);

	return token();
}

void PglPp::advance() {
	theToken = lexer()->nextToken();
}

// returns true if it is OK to pass current symbol to the parser
bool PglPp::filtered() {
	Assert(lexer());
	switch (symbol()) {
		case CMNT_LINE_TOKEN:
			ignoreLineCmnt(); 
			return false;
		case CMNT_OPEN_TOKEN:
			ignoreBlockCmnt();
			return false;
		case BQW_STR_TOKEN:
			system();
			return false;
		case POUND_TOKEN: {
			// pp-directives must be the first token on a line
			if (!token().firstOnLine())
				return true;
			ppDirective();
			return false;
		}
		case _EOF_TOKEN: {
			close();
			if (theDepth)
				advance();
			return false;
		}
		default:
			return true;
	}
}

void PglPp::ignoreLineCmnt() {
	const int lineno = token().loc().lineno();
	do {
		advance();
	} while (token().loc().lineno() == lineno && symbol() != _EOF_TOKEN);
}

void PglPp::ignoreBlockCmnt() {
	TokenLoc startLoc = token().loc();
	do {
		advance();
		if (symbol() == CMNT_OPEN_TOKEN) {
			cerr << token().loc() << "warning: nested /* comments */" << endl;
			cerr << startLoc << "warning: start of a surrounding /* comment */" << endl;
			ignoreBlockCmnt();
		}
	} while (symbol() != CMNT_CLOSE_TOKEN && symbol() != _EOF_TOKEN);

	if (symbol() == CMNT_CLOSE_TOKEN)
		advance();
}

void PglPp::ppdInclude() {
	const char *fname = spelling();

	if (symbol() != DQW_STR_TOKEN)
		cerr << token().loc() << "`#include' expects \"FILENAME\", got `" << fname << "'" << endl << xexit;

	open(fname);
	advance();
}

void PglPp::ppDirective() {
	LexToken pound = theToken;

	advance();
	LexToken name = theToken;

	if (name.id() != ID_TOKEN)
		cerr << token().loc() << "malformed preprocessor directive near `" << name.spell() << "'" << endl << xexit;

	if (!pound.loc().sameLine(name.loc()))
		return; // empty pp directive is ignored

	advance();

	if (!strcmp(name.spell(), "include"))
		ppdInclude();
	else
		cerr << name.loc() << "unknown preprocessor #directive `" << name.spell() << "'" << endl << xexit;
}

void PglPp::system() {
	const String cmd = spelling();

	if (cmd.len()) { // ignore empty commands?
		popen(cmd);
		advance();
	}
}

void PglPp::open(const String &fname) {
	if (fname == "-") {
		open(*new istream(cin.rdbuf()), fname);
		return;
	}

	for (int i = -1; i < TheDirs.count(); ++i) {
		const String fn = i >= 0 ? *TheDirs[i] + "/" + fname : fname;
		if (istream &is = *new ifstream(fn)) {
			open(is, fn);
			return;
		}
	}

	printLexers(cerr);
	cerr << token().loc() << "cannot open `" << fname << "';"
		<< " tried " << 1+TheDirs.count() << " location(s):"
		<< endl;
	for (int i = -1; i < TheDirs.count(); ++i) {
		const String fn = i >= 0 ? *TheDirs[i] + "/" + fname : fname;
		cerr << "\t " << fn << endl;
	}
	cerr << xexit;
}

void PglPp::popen(const String &cmd) {
	// unfortunately, pipes are not yet stadard in iostream lib
	// we emulate similar functionality by using stdio calls

	strstream &sbuf = *new strstream;

	// open the pipe to cmd
	FILE *fp = ::popen(cmd, "r");

	if (fp) {
		ifstream is(fileno(fp));

		// copy everything
		is.get(*sbuf.rdbuf(), (char)EOF);
	}

	// close the pipe
	if (!fp || pclose(fp) != 0)
		cerr << token().loc() << "command `" << cmd << "' probably failed." << endl << xexit;

	// use buffered input
	const String h = cmd.len() > 13 ? cmd(0,10) + "..." : cmd;
	open(sbuf, "`" + h + "`");
}

void PglPp::open(istream &is, const String &fname) {
	if (theDepth >= (int)(sizeof(theLexers)/sizeof(*theLexers))) {
		printLexers(cerr);
		cerr << token().loc() << "nesting is too deep in preprocessor" << endl << xexit;
	}

	theLexers[theDepth++] = new PglLexer(is, fname);
}

void PglPp::close() {
	if (theDepth <= 0 || !lexer())
		cerr << here << "internal error, lost lexer?" << endl << xabort;
	istream &is = lexer()->source();
	delete lexer();
	delete &is;
	theLexers[--theDepth] = 0;
}

void PglPp::printLexers(ostream &os) const {
	for (int i = theDepth-1; i > 0; --i) {
		os << theLexers[i-1]->token().loc() << "inclusion of `" << theLexers[i]->fname() << "'" << endl;
	}
}
