
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "xstd/Clock.h"
#include "base/polyLogCats.h"
#include "runtime/PortMgr.h"
#include "runtime/ErrorMgr.h"
#include "runtime/StatPhase.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/LogComment.h"
#include "runtime/Connection.h"
#include "runtime/polyBcastChannels.h"
#include "runtime/globals.h"


int Connection::TheLastSeqId = 0;


/* Connection::HalfPipe */

void Connection::HalfPipe::reset() {
	Assert(!theReserv);
	theBuf.reset();
	theIOCnt = 0;
	isReady = false;
}

void Connection::HalfPipe::start(FileScanUser *u) {
	Assert(!theReserv);
	theReserv = TheFileScanner->setFD(theConn.theSock.fd(), theDir, u);
}

void Connection::HalfPipe::start(FileScanUser *u, Time timeout) {
	start(u);
	TheFileScanner->setTimeout(theReserv.fd(), timeout);
}

void Connection::HalfPipe::stop(FileScanUser *) {
	if (theReserv)
		TheFileScanner->clearRes(theReserv);
	theBuf.pack();
}


/* Connection */

Connection::Connection():
	theRd(*this, theRdBuf, dirRead),
	theWr(*this, theWrBuf, dirWrite) {
	reset();
}

void Connection::reset() {
	if (theSock)
		close();
	theAddr = NetAddr();
	theMgr = 0;
	thePortMgr = 0;
	theRd.reset();
	theWr.reset();
	theCloseKind = ConnCloseStat::ckActive;
	theLogCat = lgcAll;
	isBad = isAtEof = isLastUse = false;
	theLocPort = theRemPort = -1;

	theUseLmt = -1; // unlimited use by default
	theUseCnt = 0;
	theOpenTime = theUseStart = Time();
	theMaxIoSize = -1;

	theSeqId = ++TheLastSeqId;
}

int Connection::rport() const {
	if (!theSock)
		return -1;
	else
	if (theRemPort >= 0)
		return theRemPort;
	else
		return theRemPort = theSock.rport(); // cache the result
}

bool Connection::connect(const NetAddr &addr, const SockOpt &opt, PortMgr *aPortMgr) {
	Assert(aPortMgr);
	Assert(!theSock);
	if (!theSock.create()) {
		ReportError(errno);
		return false;
	}

	if (!setSockOpt(opt))
		return false;

	thePortMgr = aPortMgr;
	theLocPort = thePortMgr->bind(theSock);
	if (theLocPort < 0)
		return false;

	theOpenTime = TheClock;
	Broadcast(TheConnOpenChannel, this);

	return theSock.connect(theAddr = addr);
}

bool Connection::accept(Socket &s, const SockOpt &opt, bool &fatal) {
	Assert(!theSock);

	sockaddr_in addr;
	socklen_t addr_len = sizeof(addr);
	theSock = s.accept((struct sockaddr*)&addr, &addr_len);
	if (!theSock) {
		fatal = errno != EWOULDBLOCK && errno != EAGAIN && errno != EMFILE;
		if (fatal)
			ReportError(errno);
		return false;
	}

	if (!setSockOpt(opt))
		return false;

	theAddr.addr(addr.sin_addr);
	theAddr.port(ntohs(addr.sin_port));
	theLocPort = -1; // unknown, do not care;
	theOpenTime = TheClock;

	Broadcast(TheConnOpenChannel, this);

	return true;
}

Size Connection::read() {
	if (!theRdBuf.spaceSize() || !theMaxIoSize) {
		theRd.isReady = true;
		return 0;
	}
	theRd.isReady = false;

	const Size ioSz = theMaxIoSize >= 0 ?
		Min(theRdBuf.spaceSize(), theMaxIoSize) : theRdBuf.spaceSize();
	const Size sz = preIO(theRd) ?
		theSock.read(theRdBuf.space(), ioSz) : (Size)-1;
	theMaxIoSize = -1;

	//cerr << here << "read " << sz << " : {" << theRdBuf.space() << "}" << endl;

	if (sz > 0) {
		theRdBuf.appended(sz);
		TheStatPhaseMgr->noteSockRead(sz, logCat());
		return sz;
	}

	if (sz == 0) {
		isAtEof = true;
		return sz;
	}

	/* error handling, sz < 0 */
	if (errno == EWOULDBLOCK)
		return 0;
	if (ReportError(errno))
		Comment << "error reading from " << theAddr << endc;
	isBad = true;
	return sz;
}

Size Connection::write() {
	if (!theWrBuf.contSize() || !theMaxIoSize) {
		theWr.isReady = true;
		return 0;
	}
	theWr.isReady = false;

	const Size ioSz = theMaxIoSize >= 0 ?
		Min(theWrBuf.contSize(), theMaxIoSize) : theWrBuf.contSize();
	const Size sz = preIO(theWr) ?
		theSock.write(theWrBuf.content(), ioSz) : (Size)-1;
	theMaxIoSize = -1;
	
	//cerr << here << "wrote " << sz << " : {" << theWrBuf.content() << "}" << endl;

	if (sz >= 0) {
		if (sz > 0) {
			theWrBuf.consumed(sz);
			theWrBuf.pack();
		}
		TheStatPhaseMgr->noteSockWrite(sz, logCat());
		return sz;
	}

	/* error handling, sz < 0 */
	if (errno == EWOULDBLOCK)
		return 0;
	if (ReportError(errno))
		Comment << "error writing to " << theAddr << endc;
	isBad = true;
	return sz;
}

bool Connection::close() {
	if (theLocPort >= 0) {
		thePortMgr->release(theLocPort, true);
		theLocPort = -1;
	}
	if (theSock) {
		theRd.stop(0);
		theWr.stop(0);
		Broadcast(TheConnCloseChannel, this);
		if (!theSock.close()) {
			ReportError(errno);
			return false;
		}
	}
	return true;
}

// this has to be done for accepted sockets as well
// FreeBSD accept(2) man page misleads
bool Connection::setSockOpt(const SockOpt &opt) {
	return Should(theSock.blocking(false) && theSock.configure(opt));
}

bool Connection::preIO(HalfPipe &ioPipe) {
	if (theRd.theIOCnt + theWr.theIOCnt == 0) {
		// check for connect/accept errors
		if (theSock.errNo())
			return false;
		Broadcast(TheConnEstChannel, this);
	}
	ioPipe.theIOCnt++;
	return true;
}

bool Connection::reusable() const {
	if (exhausted() || isLastUse || isBad)
		return false;
	return theUseLmt < 0 || theUseCnt < theUseLmt;
}

// prepare for next use if possible
// note: the notion of ``use'' is caller defined
void Connection::use() {
	Assert(reusable());
	// first user takes the hit of the opening delay
	theUseStart = theUseCnt++ ? (Time)TheClock : theOpenTime;
}
