
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "polygraph.h"

#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <arpa/inet.h>

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#include "base/CmdLine.h"
#include "base/opts.h"

#include "PglNetAddrRange.h"
#include "getIfAddrs.h"
#include "gadgets.h"


class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,     "help",         "list of options"),
			theIfName(this,      "if <str>",     "interface name"),
			theAliases(this,     "aliases <str>","ip(s) or ip range(s) to set as aliases")
			{}

		virtual bool validate() const;

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

	public:
		HelpOpt theHelpOpt;
		StrOpt theIfName;
		StrArrOpt theAliases;
};


struct ifaliasreq_in {
	char    ifra_name[IFNAMSIZ];   /* if name, e.g. "en0" */
	struct  sockaddr_in ifra_addr;
	struct  sockaddr_in ifra_broadaddr;
	struct  sockaddr_in ifra_mask;
};


static MyOpts TheOpts;
static Array<in_addr>  TheOldAliases;
static NetAddr ThePrimaryAddr;
//static NetAddr ThePrimaryBcastAddr;


/* MyOpt */

ostream &MyOpts::printAnonym(ostream &os) const {
	return os << "[interface_name] [alias] ...";
}

bool MyOpts::parseAnonym(const Array<const char *> &opts) {
	if (!opts.count())
		return true;

	int aliasStart = 0;
	if (!isdigit(*opts[0])) {
		theIfName.val(opts[0]);
		aliasStart = 1;
	}

	for (int i = aliasStart; i < opts.count(); ++i)
		theAliases.addItem(opts[i]);

	return true;
}

bool MyOpts::validate() const {
	if (!theIfName)
		cerr << "interface name is not specified" << endl;
	else
		return true;
	return false;
}

static
void configureLogs(int prec) {
	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

static
void configure() {
	configureLogs(2);
}

static
void getPrimaryAddress() {
	sockaddr_in addr;
	Socket s;
	Must(s.create(AF_INET, SOCK_DGRAM, 0));
	Must(s.getIfAddr(TheOpts.theIfName, (sockaddr&)addr));
	ThePrimaryAddr = NetAddr(addr);
	//Must(s.getIfBcastAddr(TheOpts.theIfName, (sockaddr&)addr));
	//ThePrimaryBcastAddr = NetAddr(addr);
	s.close();

	ThePrimaryAddr.port(-1);
	//ThePrimaryBcastAddr.port(-1);
	//clog << TheOpts.theIfName << ": primary address is " << ThePrimaryAddr << " broadcast " << ThePrimaryBcastAddr << endl;
	clog << TheOpts.theIfName << ": primary address is " << ThePrimaryAddr << endl;
}

static
void getOldAliases() {
	Must(GetIfAddrs(TheOldAliases, TheOpts.theIfName));

	// remove primary address from the aliases array
	bool found = false;
	for (int i = 0; !found && i < TheOldAliases.count(); ++i) {
		if (TheOldAliases[i] == ThePrimaryAddr.addrN()) {
			TheOldAliases[i] = TheOldAliases.last();
			TheOldAliases.pop();
			found = true;
		}
	}

	if (!found)
		clog << TheOpts.theIfName << ": hmm.. primary address is not"
			<< " in the interface alias list." << endl;

	clog << TheOpts.theIfName << ": found "
		<< TheOldAliases.count() << " old aliases." << endl;
}

static
void delOldAliases() {
	Socket s;
	Must(s.create(AF_INET, SOCK_DGRAM, 0));
	// reverse order seems to be important on Linux where
	// one cannot delete :1 after deleting :0
	for (int i = TheOldAliases.count() - 1; i >= 0; --i) {
		struct ifaliasreq_in r;
		memset(&r, 0, sizeof(r));
		strncpy(r.ifra_name, TheOpts.theIfName, sizeof(r.ifra_name));
// XXX: remove all these defines
#ifdef HAVE_SA_LEN
		r.ifra_addr.sin_len = sizeof(r.ifra_addr);
#endif
		r.ifra_addr.sin_family = AF_INET;
		r.ifra_addr.sin_port = 0;
		r.ifra_addr.sin_addr = TheOldAliases[i];
		Must(s.delIfAddr((ifaliasreq&)r, i));
	}
	s.close();
	clog << TheOpts.theIfName << ": deleted "
		<< TheOldAliases.count() << " old aliases." << endl;
}

static
void addNewAliases(const Array<NetAddr*> &aliases, const struct in_addr &netmask) {
	Socket s;
	Must(s.create(AF_INET, SOCK_DGRAM, 0));
	for (int i = 0; i < aliases.count(); ++i) {
		struct ifaliasreq_in r;
		memset(&r, 0, sizeof(r));
		strncpy(r.ifra_name, TheOpts.theIfName, sizeof(r.ifra_name));
#ifdef HAVE_SA_LEN
		r.ifra_addr.sin_len = sizeof(r.ifra_addr);
#endif
		r.ifra_addr.sin_family = AF_INET;
		r.ifra_addr.sin_port = 0;
		r.ifra_addr.sin_addr = aliases[i]->addrN();

		// set network mask (geting old netmask is not documented?!)
		r.ifra_mask = r.ifra_addr;
		r.ifra_mask.sin_addr = netmask;

		// set broacast address
		r.ifra_broadaddr = r.ifra_addr;
		r.ifra_broadaddr.sin_addr.s_addr |= ~(unsigned int)r.ifra_mask.sin_addr.s_addr;
		//r.ifra_broadaddr.sin_addr = ThePrimaryBcastAddr.addrN();

		Must(s.addIfAddr((ifaliasreq&)r, i) || errno == EEXIST);
	}
	s.close();

	static int newCount = 0;
	newCount += aliases.count();
	clog << TheOpts.theIfName << ": created "
		<< aliases.count() << " new aliases " 
		<< "( " << newCount << " total )"
		<< endl;
}

int main(int argc, char *argv[]) {

	CmdLine cmd;
	cmd.configure(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();

	clog << TheOpts.theIfName << ": ";
	if (TheOpts.theAliases)
		TheOpts.theAliases.report(clog << "reseting aliases to ");
	else
		clog << "no new alias specified; will just delete old ones.";
	clog << endl;

	getPrimaryAddress();
	getOldAliases();
	delOldAliases();

	for (int i = 0; i < TheOpts.theAliases.val().count(); ++i) {
		const String &alias = *TheOpts.theAliases.val()[i];

		PglNetAddrRange aliasParser;
		if (!aliasParser.parse(alias)) {
			cerr << TheOpts.theIfName << ": malformed alias: `" << alias << "' alias" << endl;
			return -2;
		}
		Array<NetAddr*> aliases;
		aliasParser.toAddrs(aliases);

		struct in_addr netmask;
		aliasParser.netmask(netmask);
		addNewAliases(aliases, netmask);
	}

	return 0;
}
